import json
import os

class Option:
    def __init__(self, name, helptext):
        self.name = name
        self.helptext = helptext

class CommandTree:
    def __init__(self, node=None, helptext=None, children=None, local_flags=None):
        self.node = node
        self.help = helptext
        self.children = children if children else []
        self.local_flags = local_flags if local_flags else []

class Parser:
    def __init__(self, apiFile, root_name):
        try:
            self.json_api = apiFile
            self.schema = dict()
            self.global_flags = list()
            self.root_name = root_name
            with open(self.json_api) as api:
                self.schema = json.load(api)
            self.ast = CommandTree(root_name)  # Set to dynamic root name
            root_schema = self.schema.get(root_name)
            if root_schema is None:
                logger.error(f"{root_name} is not in the schema")
            else:
                self.ast = self.build(self.ast, root_schema)
        except Exception as ex:
            logger.error(f"Exception while initializing Parser: {ex}")


    def set_api_file(self, api_file):
        self.json_api = api_file

        with open(self.json_api) as api:
            self.schema = json.load(api)

        api_schema = self.schema.get(self.json_api.split('/')[-1].split('.')[0])

        if api_schema:
            self.ast = self.build(self.ast, api_schema)
        else:
            print(f"{self.json_api.split('/')[-1].split('.')[0]} is not in the schema")

    def get_top_level_commands(self):
        return [node.node for node in self.ast.children]

    def build(self, root, schema):
        try:
            if schema.get("subcommands") and schema["subcommands"]:
                for subcmd, childSchema in schema["subcommands"].items():
                    child = CommandTree(node=subcmd)
                    child = self.build(child, childSchema)
                    root.children.append(child)
    
            for name, desc in schema.get("options", {}).items():
                root.local_flags.append(Option(name, desc["help"]))
    
            for arg in schema.get("args", []):
                node = CommandTree(node=arg)
                root.children.append(node)
    
            root.help = schema.get("help")
            return root
        except:
            pass
    
    def parse_tokens(self, tokens):
        if len(tokens) == 1:
            return [], tokens, {"oci": self.ast.help}
        else:
            tokens.reverse()

        parsed, unparsed, _ = self.treewalk(self.ast, parsed=[], unparsed=tokens)

        current_node = self.get_node_from_tokens(parsed)
        suggestions = {child.node: child.help for child in current_node.children}
        suggestions.update({flag.name: flag.helptext for flag in current_node.local_flags})

        if not suggestions and unparsed:
            unparsed.pop()
            parsed.reverse()
            unparsed.extend(parsed)
            return self.treewalk(self.ast, parsed=[], unparsed=unparsed)
        else:
            return parsed, unparsed, suggestions

    def get_node_from_tokens(self, parsed_tokens):
        current_node = self.ast

        for token in parsed_tokens:
            if token.startswith("--") or token.startswith("-"):  # Skip options
                continue
            for child in current_node.children:
                if child.node == token:
                    current_node = child
                    break
        return current_node

    def treewalk(self, root, parsed, unparsed):
        suggestions = {}

        if not unparsed:
            return parsed, unparsed, suggestions

        token = unparsed.pop().strip()

        if root.node == token:
            parsed.append(token)

            if self.peek_for_option(unparsed):
                parsed_opts, unparsed, suggestions = self.eval_options(root, [], unparsed[:])
                if parsed_opts:
                    parsed.extend(parsed_opts)

            if unparsed and not self.peek_for_option(unparsed):
                for child in root.children:
                    parsed_subtree, unparsed, child_suggestions = self.treewalk(child, [], unparsed[:])
                    if parsed_subtree:
                        parsed.extend(parsed_subtree)
                        suggestions.update(child_suggestions)
                        break
                else:
                    for child in root.children:
                        suggestions[child.node] = child.help
        else:
            unparsed.append(token)

        return parsed, unparsed, suggestions

    def peek_for_option(self, unparsed):
        if unparsed and (unparsed[-1].startswith("--") or unparsed[-1].startswith("-")):
            return True
        return False

    def eval_options(self, root, parsed, unparsed):
        suggestions = {}

        while unparsed:
            token = unparsed.pop().strip()

            parts = token.partition('=')
            if parts[-1] != '':  # parsing for --option=value type input
                token = parts[0]

            all_flags = root.local_flags + self.global_flags
            found = False

            for flag in all_flags:
                if flag.name == token:
                    parsed.append(token)
                    found = True
                    break

            if not found:
                unparsed.append(token)
                break

        for flag in all_flags:
            suggestions[flag.name] = flag.helptext

        return parsed, unparsed, suggestions




from __future__ import print_function, unicode_literals
import sys, os

from prompt_toolkit import Application
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.layout.containers import VSplit, Window
from prompt_toolkit.layout.controls import BufferControl, FormattedTextControl
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.shortcuts import print_container
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout.dimension import D
from prompt_toolkit.styles import Style

from prompt_toolkit.layout.containers import (
    HorizontalAlign,
    HSplit,
    VerticalAlign,
    VSplit,
    Window,
    WindowAlign,
)

from prompt_toolkit.widgets import (
    Frame,
    Label,
    TextArea,
)

buffer1 = Buffer()

def is_non_zero_file(fpath):
    return os.path.isfile(fpath) and os.path.getsize(fpath) > 0

tip = 'TIP: resource completion coming soon!'
version = '1.0.0'
text = f"Purpose: Cloud Wrapper\n{tip}"
title = HTML("""<b>GOAT INTERFACE</b>""")
func_title = HTML("""<b>Hotkeys</b>""")
misc_title = HTML("""<b>Commands</b>""")
advanced_title = HTML("""<b>Advanced</b>""")
input = """\
[F9]  Toggle VI mode
[F10] Toggle Profile
[TAB] Fuzzy Complete
"""

misc_input = """\
e|exit    : exit shell
c|clear   : clear screen
h|help    : display usage
"""

advanced_input = """\
history   : shell history
!<cmd>    : run OS command
cloud     : view cloud details
"""

def getLayout():

    os.system('clear')
    print_container(
    HSplit([
        Frame(
            Window(FormattedTextControl(text), height=2, align=WindowAlign.CENTER), title=(title)),
        VSplit([
            Frame(body=Label(text=input.strip()), width=30, title=(func_title)),
            Frame(body=HSplit([Label(text=advanced_input.strip())], padding=1, padding_char=' '), width=40, title=(advanced_title)),
            Frame(body=Label(text=misc_input.rstrip()), width=30, title=(misc_title))
        ])
    ]))

if __name__ == '__main__':
    getLayout()



from __future__ import print_function, absolute_import, unicode_literals

from pygments.token import Token
from pygments.util import ClassNotFound
from pygments.styles import get_style_by_name
from prompt_toolkit.styles import Style

styles = {
    'token.app.title': 'bg:#005577 #ffffff',
    'completion-menu.completions.current': 'bg:#afb3b5 #666d6f',
    'completion-menu.completion': 'bg:#666d6f #afb3b5',
    'completion-menu.meta.current': 'bg:#afb3b5 #000000',
    'completion-menu.meta': 'bg:#afb3b5 #7f7f7f',
    'bottom-toolbar': 'bg:#ffffff #000000',
    'bottom-toolbar.off': 'bg:#222222 #696969',
    'bottom-toolbar.on': 'bg:#000000 #ffffff',
    'bottom-toolbar.search': 'noinherit bold',
    'bottom-toolbar.search.text': 'noinherit bold',
    'bottom-toolbar.system': 'noinherit bold',
    'bottom-toolbar.arg': 'noinherit bold'
}



from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.completion import WordCompleter

import os
import sys
import subprocess
import re
import click
import logging
import configparser
from prompt_toolkit.patch_stdout import patch_stdout
from prompt_toolkit import Application, PromptSession
from prompt_toolkit.history import InMemoryHistory, FileHistory
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.styles import Style
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.keys import Keys

from . import misc as misc
from goatshell.style import styles
from goatshell.completer import GoatCompleter
from goatshell.parser import Parser
from goatshell.ui import getLayout
from pygments.token import Token

logger = logging.getLogger(__name__)

# Global variables
global vi_mode_enabled
current_service = 'oci'       # You can set it to 'aws' if you prefer AWS mode initially
os.environ['AWS_PAGER'] = ''  # Disable paging in AWS by default
vi_mode_enabled = False       # Default vi mode to false on start up

# Function to print instructions
def printInstructions():
    instructions = """
    Auto-Completion Instructions:
    ----------------------------
    1. To trigger auto-completion, start with TAB or type the beginning of a command or option and press Tab.
    2. Auto-completion will suggest available commands, options, and arguments based on your input.
    3. Use the arrow keys or Tab to navigate through the suggestions.
    4. Press Enter to accept a suggestion or Esc to cancel.
    5. If an option requires a value, use --option=value instead of --option value.
    6. The prompt will change dynamically based on cloud provider interaction.
    """
    print(instructions)

# Function to save VI mode setting
def save_vi_mode_setting():
    global vi_mode_enabled
    shell_dir = os.path.expanduser("~/goat/shell/")
    if not os.path.exists(shell_dir):
        os.makedirs(shell_dir)
    config = configparser.ConfigParser()
    config['Settings'] = {'vi_mode_enabled': str(vi_mode_enabled)}
    with open(shell_dir + 'config.ini', 'w') as configfile:
        config.write(configfile)

# Function to load VI mode setting from a configuration file
def load_vi_mode_setting():
    global vi_mode_enabled
    shell_dir = os.path.expanduser("~/goat/shell/")
    config = configparser.ConfigParser()
    try:
        config.read(shell_dir + 'config.ini')
        vi_mode_enabled = config.getboolean('Settings', 'vi_mode_enabled')
    except (configparser.NoSectionError, configparser.NoOptionError, FileNotFoundError):
        # Handle exceptions or file not found gracefully
        pass
    return vi_mode_enabled

# Load the VI mode setting when the script starts
vi_mode_enabled = load_vi_mode_setting()

# Define a custom PromptSession class
class DynamicPromptSession(PromptSession):
    def __init__(self, vi_mode_enabled=True, *args, **kwargs):
        self.vi_mode_enabled = vi_mode_enabled
        super().__init__(*args, **kwargs)

    def prompt(self, *args, **kwargs):
        # Set the 'vi_mode' parameter based on 'vi_mode_enabled'
        kwargs['vi_mode'] = self.vi_mode_enabled
        return super().prompt(*args, **kwargs)

# Define key bindings class
class CustomKeyBindings(KeyBindings):
    def __init__(self, app, goatshell_instance):
        super().__init__()
        self.app = app  # Store the Application reference
        self.goatshell_instance = goatshell_instance

        @self.add(Keys.F10)
        def handle_f10(event):
            self.profile = self.goatshell_instance.get_profile(self.goatshell_instance.prefix)  # Access the get_profile method of Goatshell
            getLayout()
            self.app.invalidate()
            event.app.exit(result='re-prompt')  # Signal to reprompt.

        @self.add(Keys.F9)
        def handle_f9(event):
            self.goatshell_instance.toggle_vi_mode()
            self.app.invalidate()
            event.app.exit(result='re-prompt')  # Signal to reprompt.

# Define the main Goatshell class
class Goatshell(object):
    def __init__(self, app, completer, parser):
        getLayout()
        self.vi_mode_enabled = load_vi_mode_setting()
        self.style = Style.from_dict(styles)
        self.app = app
        self.completer = completer
        self.prefix = 'oci'
        self.current_input = None
        self.aws_profiles = misc.read_aws_profiles()
        self.oci_profiles = misc.read_oci_profiles()
        self.aws_index = 0
        self.oci_index = 0
        self.cloud_providers = ['aws','oci','ibmcloud','gcloud','goat','az','aliyun']
        self.profile = 'DEFAULT'  # Init this variable before we call the get_profile function
        self.profile = self.get_profile(self.prefix) # now update the profile
        self.key_bindings = CustomKeyBindings(self.app, self)
        shell_dir = os.path.expanduser("~/goat/shell/")
        if not os.path.exists(shell_dir):
            os.makedirs(shell_dir)
        try:
            self.history = FileHistory(os.path.join(shell_dir, "history"))
        except:
            self.history = InMemoryHistory()

        # Pass the 'vi_mode_enabled' variable when initializing DynamicPromptSession
        self.session = DynamicPromptSession(vi_mode_enabled=self.vi_mode_enabled, history=self.history, auto_suggest=AutoSuggestFromHistory(),
                                            completer=self.completer, complete_while_typing=True,
                                            enable_history_search=True, key_bindings=self.key_bindings)

        root_name, json_path = self.get_service_info()
        try:
            self.parser = Parser(json_path, root_name)
        except Exception as e:
            logger.info(f"Exception caught: {e}")
            logger.info(f"json_path = {json_path}, root_name = {root_name}")

    def toggle_vi_mode(self):
        self.vi_mode_enabled = not self.vi_mode_enabled
        save_vi_mode_setting()
        self.session = DynamicPromptSession(vi_mode_enabled=self.vi_mode_enabled, history=self.history, auto_suggest=AutoSuggestFromHistory(),
                                           completer=self.completer, complete_while_typing=True,
                                           enable_history_search=True, key_bindings=self.key_bindings)

    def get_account_or_tenancy(self, profile):
        if self.prefix == 'oci':
            # Label as account for variable consistency and get the OCID
            account = misc.get_oci_tenant(profile_name=profile).id
        elif self.prefix == 'aws':
            account = misc.get_aws_account(profile_name=profile)
        else:
            return 'UNKNOWN'
        return account

    def get_profile(self, mode):
        if mode == 'aws':
            if self.aws_profiles:
                if self.profile in self.aws_profiles:
                    self.aws_index = self.aws_profiles.index(self.profile)
                self.aws_index = (self.aws_index + 1) % len(self.aws_profiles)
                self.profile = self.aws_profiles[self.aws_index]
            else:
                self.profile = 'DEFAULT'
        elif mode == 'oci':
            if self.oci_profiles:
                if self.profile in self.oci_profiles:
                    self.oci_index = self.oci_profiles.index(self.profile)
                self.oci_index = (self.oci_index + 1) % len(self.oci_profiles)
                self.profile = self.oci_profiles[self.oci_index]
            else:
                self.profile = 'DEFAULT'
        else:
            self.profile = 'DEFAULT'
        return self.profile

    def get_service_info(self):
        json_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), f'data/{self.prefix}.json')
        global current_service  # Declare current_service as a global variable
        if current_service == 'aws':
            return 'aws', json_path
        elif current_service == 'oci':
            return 'oci', json_path
        elif current_service == 'gcloud':
            return 'gcloud', json_path
        elif current_service == 'az':
            return 'az', json_path
        elif current_service == 'goat':
            return 'goat', json_path
        elif current_service == 'aliyun':
            return 'aliyun', json_path
        elif current_service == 'ibmcloud':
            return 'ibmcloud', json_path
        else:
            return 'aws', json_path

    def display_environment(self):
        details = {}
        if self.prefix == 'oci':
            user = misc.get_oci_user(self.profile)
            tenant = misc.get_oci_tenant(self.profile)
        elif self.prefix == 'aws':
            user = misc.get_aws_user(self.profile)
            tenant = misc.get_aws_account(self.profile)
        if user and tenant:
            details["user"] = user.strip()
            details["tenant"] = tenant
        return details

    def create_toolbar(self):
        self.upper_profile = self.profile.upper()
        self.upper_prefix = self.prefix.upper()
        vi_mode_text = "ON" if self.vi_mode_enabled else "OFF"

        return HTML(f'Current Cloud: <u>{self.upper_prefix}</u>   <b>F9 VIM {vi_mode_text}</b>   <b>F10</b> Profile: <u>{self.upper_profile}</u>')

    def set_parser_and_completer(self, api_type):
        self.prefix = api_type.lower()  # Set prefix
        json_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), f'data/{api_type}.json')
        self.parser = Parser(json_path, api_type)  # Reset parser
        self.completer = GoatCompleter(self.parser)  # Reset completer
        self.session.completer = self.completer  # Reset session completer

    def process_user_input(self, user_input):
        user_input = user_input.strip()
      
        # Handle the history command
        if user_input == "history":
            history_file = os.path.expanduser("~/goat/shell/history")
            if os.path.exists(history_file):
                with open(history_file, "r") as f:
                    # Enumerate over the file lines and print them with line numbers
                    for index, line in enumerate(f, 1):
                        print(f"{index}  {line.strip()}")
                return ""  # or return None, so nothing further is executed for this command
            else:
                print("History file not found.")
                return ""
        elif user_input == "cloud":
            details = self.display_environment()
            details_str = f"Cloud Provider: {self.prefix.upper()}\n"
            for key, value in details.items():
                details_str += f"{key.upper()}: {value}\n"
            print()
            print(details_str)
            return ""
            
        if user_input.startswith("!"):
            user_input = user_input[1:]
        else:
            tokens = user_input.split(' ')
            first_token = tokens[0]
            last_token = tokens[-1]
            last_but_one_token = tokens[-2] if len(tokens) > 1 else None

            # Process known commands
            if first_token == 'oci':
                user_input = self.process_oci_input(user_input, first_token, last_token, last_but_one_token)
            elif first_token == 'aws':
                user_input = self.process_aws_input(user_input, first_token, last_token, last_but_one_token)

            # Common processing for aws and oci
            if first_token in ['aws', 'oci'] and '--profile' not in user_input:
                user_input = user_input + ' --profile ' + self.profile

                user_input = user_input + ' --profile ' + self.profile

            if '-o' in user_input and 'json' in user_input:
                user_input += ' | pygmentize -l json'

            if first_token.lower() not in self.cloud_providers:
                if first_token.lower() not in ['help','h','c','clear','e','exit']:
                    print("INFO: please select a cloud provider as the first command, or precede OS commands with an !")
                    user_input = '' 
        return user_input

    def process_oci_input(self, user_input, first_token, last_token, last_but_one_token):
        if self.profile not in self.oci_profiles:
            self.profile = self.get_profile(first_token.lower())

        if last_but_one_token in ['--compartment-id', '--tenancy-id'] and not last_token.startswith('ocid'):
            try:
                OCID = self.get_account_or_tenancy(self.profile)
                user_input += f' {OCID}'
            except:
                pass

        if last_but_one_token == '--user-id' and not last_token.startswith('ocid'):
            try:
                OCID = misc.get_oci_user(self.profile)
                user_input += f' {OCID}'
            except:
                pass

        return user_input

    def process_aws_input(self, user_input, first_token, last_token, last_but_one_token):
        if self.profile not in self.aws_profiles:
            self.profile = self.get_profile(first_token.lower())

        if last_but_one_token == '--user-name' and not last_token:
            try:
                USER = misc.get_aws_user(self.profile)
                user_input += f' {USER}'
            except:
                pass

        return user_input

    def get_current_context(self):
        return {
            'cloud_provider': self.prefix,
            'profile': self.profile
        }

    def generate_prompt(self):
        context = self.get_current_context()
        return HTML(f'[<b><u>{context["cloud_provider"]}</u></b>:<b><u>{context["profile"]}</u></b>]> ')

    def run_cli(self):
        global current_service
        logger.info("running goat event loop")
        while True:
            prompt = self.generate_prompt()
            try:
                user_input = self.session.prompt(prompt,
                                                 style=self.style,
                                                 completer=self.completer,
                                                 complete_while_typing=True,
                                                 bottom_toolbar=self.create_toolbar())
            except (EOFError, KeyboardInterrupt):
                sys.exit()

            if user_input == 're-prompt':
                user_input = ''
                continue
            api_type = user_input.split(' ')[0]

            if api_type.lower() != self.prefix:
                if api_type.lower() in self.cloud_providers:
                    self.set_parser_and_completer(api_type.lower())

            user_input = re.sub(r'[-]{3,}', '--', user_input)
            if user_input == "clear" or user_input == 'c':
                user_input = 'clear'
                click.clear()
            elif user_input == "exit" or user_input == 'e':
                user_input = 'exit'
                sys.exit()
            elif user_input == "help" or user_input == 'h':
                getLayout()
                printInstructions()
                continue

            user_input = self.process_user_input(user_input)
            p = subprocess.Popen(user_input, shell=True)
            p.communicate()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, absolute_import, unicode_literals
from goatshell.goatshell import Goatshell
from goatshell.completer import GoatCompleter
from prompt_toolkit.styles import Style
from prompt_toolkit.application import Application
from goatshell.style import styles
import os

from goatshell.parser import Parser  # Import the Parser class from the appropriate location
from goatshell.completer import GoatCompleter  # Import the GoatCompleter class from the appropriate location

def cli():
    # Default to the 'oci' service
    service = 'oci'

    # Construct the path to the JSON file based on the selected service
    json_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), f'data/{service}.json')

    # Create a Parser instance
    parser = Parser(json_path, service)
    completer = GoatCompleter(parser)

    # Initialize and run Goatshell
    app = Application()
    goatshell = Goatshell(app, completer, parser)
    goatshell.run_cli()

if __name__ == '__main__':
    cli()

